/*
 * Author	: Hatim Kanchwala
 * Email	: hatim@hatimak.me
 */

#include <stdio.h>
#include <stdint.h>

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#define LEN_RANGES 32

#define GIGADEVICE_GD25Q64		0x4017
#define GIGADEVICE_GD25VQ41B	0x4213
#define GIGADEVICE_GD25VQ21B	0x4212
#define GIGADEVICE_GD25VQ80C	0x4214
#define WINBOND_W25X80			0x3014
#define WINBOND_W25Q64			0x6071
#define WINBOND_EN25Q32			0x3316
#define WINBOND_EN25Q128		0x3018

// Temporary block for demonstrational purposes
#ifdef SET_CMP
#define CMP_VALUE 1
#else
#define CMP_VALUE 0
#endif
const unsigned char cmp_value = CMP_VALUE;

enum bit_state {
	X 	= -1,
	OFF = 0,
	ON 	= 1,
};

enum status_register_bit {
	INVALID_BIT = 0, RESV, WIP, WEL, SRP0, SRP1, WP, BP0, BP1, BP2, BP3, BP4,
	SEC, TB, CMP, LB3, LB2, LB1, WPS, QE, SUS1, SUS2, DRV0, DRV1, RST, HPF
};

enum wp_mode {
	WP_MODE_UNKNOWN = -1,
	WP_MODE_HARDWARE,
	WP_MODE_POWER_CYCLE,
	WP_MODE_PERMANENT,
};

enum status_register_num { SR1 = 0, SR2, SR3 };

const char *status_register_bit_name[] = {
	"", "RESV", "WIP/BUSY", "WEL", "SRP0", "SRP1", "WPDIS", "BP0", "BP1", "BP2", "BP3", "BP4", "SEC", 
	"TB", "CMP", "LB3", "LB2", "LB1", "WPS", "QE", "SUS1", "SUS2", "DRV0", "DRV1", "RST", "HPF"
};

const char *bit_string_8[8]		= { "0 0 0", "0 0 1", "0 1 0", "0 1 1", "1 0 0", "1 0 1", "1 1 0", "1 1 1" };
const char *bit_string_16[16]	= { "0 0 0 0", "0 0 0 1", "0 0 1 0", "0 0 1 1", "0 1 0 0", "0 1 0 1", "0 1 1 0", "0 1 1 1",
									"1 0 0 0", "1 0 0 1", "1 0 1 0", "1 0 1 1", "1 1 0 0", "1 1 0 1", "1 1 1 0", "1 1 1 1" };
const char *bit_string_32[32]	= { "0 0 0 0 0", "0 0 0 0 1", "0 0 0 1 0", "0 0 0 1 1", "0 0 1 0 0", "0 0 1 0 1", "0 0 1 1 0", "0 0 1 1 1",
									"0 1 0 0 0", "0 1 0 0 1", "0 1 0 1 0", "0 1 0 1 1", "0 1 1 0 0", "0 1 1 0 1", "0 1 1 1 0", "0 1 1 1 1", 
									"1 0 0 0 0", "1 0 0 0 1", "1 0 0 1 0", "1 0 0 1 1", "1 0 1 0 0", "1 0 1 0 1", "1 0 1 1 0", "1 0 1 1 1",
									"1 1 0 0 0", "1 1 0 0 1", "1 1 0 1 0", "1 1 0 1 1", "1 1 1 0 0", "1 1 1 0 1", "1 1 1 1 0", "1 1 1 1 1" };

const uint16_t sec_block_multiples_64[8]	= { 0, 1, 1, 1, 0, 1, 1, 1 };
const uint16_t sec_block_multiples_128[8] 	= { 0, 2, 1, 1, 0, 2, 1, 1 };
const uint16_t sec_block_multiples_256[8] 	= { 0, 4, 2, 1, 0, 4, 2, 1 };
const uint16_t sec_block_multiples_512[8]	= { 0, 8, 4, 2, 1, 1, 1, 1 };
const uint16_t sec_block_multiples_1024[8] 	= { 0, 16, 8, 4, 2, 1, 1, 1 };
const uint16_t sec_block_multiples_2048[8]	= { 0, 32, 16, 8, 4, 2, 1, 1 };
const uint16_t sec_block_multiples_4096[8]	= { 0, 64, 32, 16, 8, 4, 2, 1 };

struct range {
	uint32_t start;
	uint32_t len;	/* in kB */
};

struct flashchip;

struct status_register {
	struct status_register_layout {
		enum status_register_bit bits[8];
	} layout[4];

	uint8_t (*read_status)(const struct flashchip *flashchip, enum status_register_num);
	int (*write_status)(const struct flashchip *flashchip, enum status_register_num, int status);
};

struct wp {
	struct range *ranges;

	/* Return pointer to WP table (CMP=0) */
	struct range* (*ranges_table)(const struct flashchip *flashchip);

	/* Given WP bits' configuration, return the corresponding range */
	struct range* (*bp_to_range)(const struct flashchip *flashchip, unsigned char bp_config);

	/* Given WP range, return WP bits' configuration (or -1, if does not exist) */
	char (*range_to_bp)(const struct flashchip *flashchip, uint32_t start, uint32_t len);

	/* Given WP bits' configuration, set status register */
	int (*bp_to_status)(const struct flashchip *flashchip, unsigned char bp_config);

	/* Get WP mode */
	enum wp_mode (*get_wp_mode)(const struct flashchip *flashchip);

	/* Set WP mode */
	int (*set_wp_mode)(const struct flashchip *flashchip);

	/* Print WP table */
	int (*list_table)(const struct flashchip *flashchip);

	/* Get value of CMP bit from chip */
	unsigned char (*get_cmp)(const struct flashchip *flashchip);

	/* Set value of CMP bit to status */
	int (*set_cmp)(const struct flashchip *flashchip, unsigned char cmp);
};

struct flashchip {
	uint32_t model_id;
	const char *name;
	unsigned int total_size;

	struct status_register *status_register;

	struct wp *wp;
};

static uint32_t wp_table_id;
static uint32_t cmp_wp_table_id;
static struct range *wp_table;
static struct range *cmp_wp_table;

int pos_bit_generic(const struct flashchip *flashchip, enum status_register_bit bit) {
	int result = -1;

	for (int i = 0; flashchip->status_register->layout[i].bits[0] != INVALID_BIT; i++) {
		for (int j = 0; j < 8; j++) {
			if (flashchip->status_register->layout[i].bits[j] == bit) {
				result = (8 * i) + j;
				break;
			}
		}
	}

	return result;
}

struct range *sec_block_pattern_range_generic(const struct flashchip *flashchip, char cmp) {
	static struct range ranges[LEN_RANGES];

	unsigned char convention = 1;
	uint32_t top = flashchip->total_size - 1;
	uint16_t *sec_block_multiples;

	switch (flashchip->total_size) {
	case 128:
		sec_block_multiples = sec_block_multiples_128;
		break;
	case 256:
		sec_block_multiples = sec_block_multiples_256;
		break;
	case 512:
		sec_block_multiples = sec_block_multiples_512;
		break;
	case 1024:
		sec_block_multiples = sec_block_multiples_1024;
		convention = 0;
		break;
	case 2 * 1024:
		sec_block_multiples = sec_block_multiples_2048;
		convention = 0;
		break;
	case 4 * 1024:
	case 8 * 1024:
	case 16 * 1024:
		sec_block_multiples = sec_block_multiples_4096;
		break;
	}

	/* SEC, TB = (0, 0) Uppper */
	for (int i = 0; i < 8; i++) {
		if (sec_block_multiples[i]) {
			ranges[i].start = (flashchip->total_size - (flashchip->total_size / sec_block_multiples[i])) * 1024;
			ranges[i].len	= flashchip->total_size / sec_block_multiples[i];
		} else {
			ranges[i].start = 0x000000;
			ranges[i].len 	= 0;
		}
	}

	if (!(pos_bit_generic(flashchip, TB) != -1 || pos_bit_generic(flashchip, BP3) != -1))
		return ranges;

	/* SEC, TB = (0, 1) Bottom */
	for (int i = 0; i < 8; i++) {
		if (sec_block_multiples[i]) {
			ranges[(0x01 << 3) | i].start 	= 0x000000;
			ranges[(0x01 << 3) | i].len	= flashchip->total_size / sec_block_multiples[i];
		} else {
			ranges[(0x01 << 3) | i].start 	= 0x000000;
			ranges[(0x01 << 3) | i].len 	= 0;
		}
	}

	if (!(pos_bit_generic(flashchip, SEC) != -1 || pos_bit_generic(flashchip, BP4) != -1))
		return ranges;

	/* SEC, TB = (1, 0) Uppper */
	ranges[(0x02 << 3) | 0x00].start = 0x000000;
	ranges[(0x02 << 3) | 0x00].len = 0;
	ranges[(0x02 << 3) | 0x01].start = (flashchip->total_size - 4) * 1024;
	ranges[(0x02 << 3) | 0x01].len = 4;
	ranges[(0x02 << 3) | 0x02].start = (flashchip->total_size - 8) * 1024;
	ranges[(0x02 << 3) | 0x02].len = 8;
	ranges[(0x02 << 3) | 0x03].start = (flashchip->total_size - 16) * 1024;
	ranges[(0x02 << 3) | 0x03].len = 16;
	ranges[(0x02 << 3) | 0x04].start = (flashchip->total_size - 32) * 1024;
	ranges[(0x02 << 3) | 0x04].len = 32;
	ranges[(0x02 << 3) | 0x05].start = (flashchip->total_size - 32) * 1024;
	ranges[(0x02 << 3) | 0x05].len = 32;
	if (convention) {
		ranges[(0x02 << 3) | 0x06].start = (flashchip->total_size - 32) * 1024;
		ranges[(0x02 << 3) | 0x06].len = 32;
	} else {
		ranges[(0x02 << 3) | 0x06].start = 0x000000;
		ranges[(0x02 << 3) | 0x06].len = flashchip->total_size;
	}
	ranges[(0x02 << 3) | 0x07].start = 0x000000;
	ranges[(0x02 << 3) | 0x07].len = flashchip->total_size;

	/* SEC, TB = (1, 1) Lower */
	ranges[(0x03 << 3) | 0x00].start = 0x000000;
	ranges[(0x03 << 3) | 0x00].len = 0;
	ranges[(0x03 << 3) | 0x01].start = 0x000000;
	ranges[(0x03 << 3) | 0x01].len = 4;
	ranges[(0x03 << 3) | 0x02].start = 0x000000;
	ranges[(0x03 << 3) | 0x02].len = 8;
	ranges[(0x03 << 3) | 0x03].start = 0x000000;
	ranges[(0x03 << 3) | 0x03].len = 16;
	ranges[(0x03 << 3) | 0x04].start = 0x000000;
	ranges[(0x03 << 3) | 0x04].len = 32;
	ranges[(0x03 << 3) | 0x05].start = 0x000000;
	ranges[(0x03 << 3) | 0x05].len = 32;
	ranges[(0x02 << 3) | 0x06].start = 0x000000;
	if (convention) {
		ranges[(0x02 << 3) | 0x06].len = 32;
	} else {
		ranges[(0x02 << 3) | 0x06].len = flashchip->total_size;
	}
	ranges[(0x03 << 3) | 0x07].start = 0x000000;
	ranges[(0x03 << 3) | 0x07].len = flashchip->total_size;

	if (cmp == ON) {
		for (int i = 0; i < LEN_RANGES; i++) {
			if (ranges[i].len == 0) {
				ranges[i].len = flashchip->total_size;
			} else if (ranges[i].len == flashchip->total_size) {
				ranges[i].len = 0;
			} else {
				if (ranges[i].start == 0x000000)
					ranges[i].start = ranges[i].len * 1024;
				else
					ranges[i].start = 0x000000;
				ranges[i].len = flashchip->total_size - ranges[i].len;
			}
		}
	}

	return ranges;
}

unsigned char bp_bitfield_generic(const struct flashchip *flashchip) {
	unsigned char bitfield = 0x00;
	enum status_register_bit bits[] = { BP0, BP1, BP2, BP3, BP4, SEC, TB };

	for (int i = 0; i < ARRAY_SIZE(bits); i++) {
		if (pos_bit_generic(flashchip, bits[i]) != -1)
			bitfield |= 1 << pos_bit_generic(flashchip, bits[i]);
	}

	return bitfield;
}

int list_table_generic(const struct flashchip *flashchip) {
	unsigned char bitfield = bp_bitfield_generic(flashchip) >> pos_bit_generic(flashchip, BP0);
	int limit = 0;
	for (int i = 0; i < 8; i++)
		if (bitfield & (1 << i))
			limit++;
	limit = 1 << limit;

	char **bit_string = (limit == 32) ? bit_string_32 : (limit == 16) ? bit_string_16 : bit_string_8;
	struct range *table = flashchip->wp->ranges_table(flashchip);

	for (int i = 0; i < limit; i++) {
		printf("%-8s\t0x%06x\t0x%06x\n",
			bit_string[i],
			(table[i].len) ? table[i].start : 0,
			(table[i].len) ? ((table[i].len * 1024) + table[i].start - 1) : 0);
	}

	return 0;
}

struct range *bp_to_range_generic(const struct flashchip *flashchip, unsigned char bp_config) {
	return &(flashchip->wp->ranges_table(flashchip)[bp_config]);
}

char range_to_bp_generic(const struct flashchip *flashchip, uint32_t start, uint32_t len) {
	struct range *table = flashchip->wp->ranges_table(flashchip);

	for (char i = 0; i < LEN_RANGES; i++) {
		if (table[i].start == start && table[i].len == len)
			return i;
	}

	return -1;
}

unsigned char get_cmp_generic(const struct flashchip *flashchip) {
	int cmp_pos = pos_bit_generic(flashchip, CMP);
	enum status_register_num sr = cmp_pos / 8;
	uint8_t status = flashchip->status_register->read_status(flashchip, sr);
	return (status & (1 << (sr * 8))) >> (sr * 8);
};

int set_cmp_generic(const struct flashchip *flashchip, unsigned char cmp) {
	int cmp_pos = pos_bit_generic(flashchip, CMP);
	enum status_register_num sr = cmp_pos / 8;
	uint8_t status = flashchip->status_register->read_status(flashchip, sr);
	status |= 1 << (sr * 8);
	return flashchip->status_register->write_status(flashchip, sr, status);
};

struct range *table_dyn_generic(const struct flashchip *flashchip) {
	// This is a temporary arrangement, actually value of CMP will come from RDSR1/2/3
	if (cmp_value) {
		if (cmp_wp_table_id != flashchip->model_id) {
			cmp_wp_table_id = flashchip->model_id;
			cmp_wp_table = sec_block_pattern_range_generic(flashchip, 1);
		}
		return cmp_wp_table;
	} else {
		if (wp_table_id != flashchip->model_id) {
			wp_table_id = flashchip->model_id;
			wp_table = sec_block_pattern_range_generic(flashchip, 0);
		}
		return wp_table;
	}
}

struct range *w25x_table_dyn(const struct flashchip *flashchip) {
	// This is a temporary arrangement, actually value of CMP will come from RDSR1/2/3
	if (wp_table_id != flashchip->model_id) {
		wp_table_id = flashchip->model_id;
		wp_table = sec_block_pattern_range_generic(flashchip, 0);
	}
	return wp_table;
}

struct range *table_man_generic(const struct flashchip *flashchip) {
	return flashchip->wp->ranges;
}

struct status_register gd25q64_statreg = {
	.layout = {
		{ WIP, WEL, BP0, BP1, BP2, BP3, BP4, SRP0 },
		{ SRP1, QE, SUS2, LB1, LB2, LB3, CMP, SUS1 },
		{ RESV, RESV, RESV, RESV, HPF, DRV0, DRV1, RESV },
	},
};

struct status_register gd25vq_statreg = {
	.layout = {
		{ WIP, WEL, BP0, BP1, BP2, BP3, BP4, SRP0 },
		{ SRP1, QE, HPF, LB1, LB2, LB3, CMP, SUS1 },
	},
};

struct status_register gd25vq16c_statreg = {
	.layout = {
		{ WIP, WEL, BP0, BP1, BP2, BP3, BP4, SRP0 },
		{ SRP1, QE, LB1, RESV, RESV, HPF, CMP, SUS1 },
	},
};

struct status_register w25x_statreg = {
	.layout = {
		{ WIP, WEL, BP0, BP1, BP2, TB, RESV, SRP0 },
	},
};

struct status_register w25q64_statreg = {
	.layout = {
		{ WIP, WEL, BP0, BP1, BP2, TB, SEC, SRP0 },
		{ SRP1, QE, RESV, LB1, LB2, LB3, CMP, SUS1 },
	},
};

struct status_register en25q_statreg = {
	.layout = {
		{ WIP, WEL, BP0, BP1, BP2, RESV, RESV, SRP0 },
	},
};

struct status_register en25q128_statreg = {
	.layout = {
		{ WIP, WEL, BP0, BP1, BP2, BP3, WP, SRP0 },
	},
};

struct wp gd25_wp = {
	.ranges_table	= table_dyn_generic,
	.bp_to_range	= bp_to_range_generic,
	.range_to_bp	= range_to_bp_generic,
	.list_table		= list_table_generic,
	.get_cmp		= get_cmp_generic,
	.set_cmp		= set_cmp_generic,
};

struct wp w25x_wp = {
	.ranges_table	= w25x_table_dyn,
	.bp_to_range	= bp_to_range_generic,
	.range_to_bp	= range_to_bp_generic,
	.list_table		= list_table_generic,
	.get_cmp		= get_cmp_generic,
	.set_cmp		= set_cmp_generic,
};

struct range en25q128_ranges[LEN_RANGES] = {
	{ 0x000000, 0 },
	{ 0x000000, 16320 },
	{ 0x000000, 16256 },
	{ 0x000000, 16128 },
	{ 0x000000, 15872 },
	{ 0x000000, 15360 },
	{ 0x000000, 14336 },
	{ 0x000000, 16384 },
	{ 0x000000, 0 },
	{ 0x010000, 16320 },
	{ 0x020000, 16256 },
	{ 0x040000, 16128 },
	{ 0x080000, 15872 },
	{ 0x100000, 15360 },
	{ 0x200000, 14336 },
	{ 0x000000, 16384 },
};

struct wp en25q128_wp = {
	.ranges			= en25q128_ranges,
	.ranges_table	= table_man_generic,
	.bp_to_range	= bp_to_range_generic,
	.range_to_bp	= range_to_bp_generic,
	.list_table		= list_table_generic,
};

const struct flashchip gd25q64 = {
	.model_id = GIGADEVICE_GD25Q64,
	.name = "GIGADEVICE_GD25Q64",
	.total_size	= 8 * 1024,
	.status_register = &gd25q64_statreg,
	.wp = &gd25_wp,	
};

const struct flashchip gd25vq41b = {
	.model_id = GIGADEVICE_GD25VQ41B,
	.name = "GIGADEVICE_GD25VQ41B",
	.total_size	= 512,
	.status_register = &gd25vq_statreg,
	.wp = &gd25_wp,	
};

const struct flashchip gd25vq21b = {
	.model_id = GIGADEVICE_GD25VQ21B,
	.name = "GIGADEVICE_GD25VQ21B",
	.total_size	= 256,
	.status_register = &gd25vq_statreg,
	.wp = &gd25_wp,	
};

const struct flashchip gd25vq16c = {
	.model_id = GIGADEVICE_GD25VQ21B,
	.name = "GIGADEVICE_GD25VQ21B",
	.total_size	= 2 * 1024,
	.status_register = &gd25vq16c_statreg,
	.wp = &gd25_wp,	
};

const struct flashchip gd25vq80c = {
	.model_id = GIGADEVICE_GD25VQ80C,
	.name = "GIGADEVICE_GD25VQ80C",
	.total_size	= 1024,
	.status_register = &gd25vq16c_statreg,
	.wp = &gd25_wp,	
};

const struct flashchip w25x80 = {
	.model_id = WINBOND_W25X80,
	.name = "WINBOND_W25X80",
	.total_size	= 1024,
	.status_register = &w25x_statreg,
	.wp = &w25x_wp,
};

const struct flashchip w25q64 = {
	.model_id = WINBOND_W25Q64,
	.name = "WINBOND_W25Q64",
	.total_size	= 8 * 1024,
	.status_register = &w25q64_statreg,
	.wp = &gd25_wp,
};

const struct flashchip en25q32 = {
	.model_id = WINBOND_EN25Q32,
	.name = "WINBOND_EN25Q32",
	.total_size	= 4 * 1024,
	.status_register = &en25q_statreg,
	.wp = &gd25_wp,
};

const struct flashchip en25q128 = {
	.model_id = WINBOND_EN25Q128,
	.name = "WINBOND_EN25Q128",
	.total_size	= 16 * 1024,
	.status_register = &en25q128_statreg,
	.wp = &en25q128_wp,
};

void test(const struct flashchip *flashchip) {
	printf("%s\n", flashchip->name);
	flashchip->wp->list_table(flashchip);
	printf("{ 0x%06x, %2d }\n", (flashchip->wp->bp_to_range(flashchip, 0x06))->start, (flashchip->wp->bp_to_range(flashchip, 0x06))->len);
	printf("%s\n", (flashchip->wp->range_to_bp(flashchip, 0x780000, 512) == -1) ? "UNKNOWN_RANGE" : bit_string_32[flashchip->wp->range_to_bp(flashchip, 0x780000, 512)]);
	printf("\n");
}

int main(int argc, const char **argv) {
	test(&gd25vq16c);
	test(&gd25vq80c);
	test(&gd25vq21b);
	test(&gd25vq41b);
	test(&gd25q64);
	test(&w25x80);
	test(&w25q64);
	test(&en25q32);
	test(&en25q128);

	return 0;
}